#define detect 3        //детектор нуля
#define triac 5        //тиристор
#include <PID_v1.h> //Библиотека ПИД http://playground.arduino.cc/Code/PIDLibrary

#include <Wire.h>
#include <LiquidCrystal_I2C.h> //https://github.com/marcoschwartz/LiquidCrystal_I2C

LiquidCrystal_I2C lcd(0x3F, 16, 2);

#define button1 8 // кнопка 1
#define button2 9 // кнопка 2

String Var;
String T1, T2;

volatile unsigned long Iism = 0; //мгновенные значения тока
volatile int cntr = 0; // счетчик в обработчике прерывания
volatile unsigned long Isumm = 0; // переменная для хранения суммы квадратов

double real_I = 0; //переменная расчета RMS тока
double sqrtUsum = 0; //переменная расчете квадратного корня
double ust_I = 0;//установленный ток
unsigned long previousMillis = 0;
unsigned long displayMillis = 0;
unsigned long reg_time = 0;

double angle_I = 625;
int t = 0;
int zero = 0;
int I_offset = 0;

int indicator_I;
unsigned long showTime = 0; //время обновления экрана
double buffer_ust_I = 0;
int button1_press_time = 0;
unsigned long time1_press = 0;
byte button1_press = 0;
int button2_press_time = 0;
unsigned long time2_press = 0;
byte button2_press = 0;
unsigned long button_timeout = 0;

double aggKp = 8, aggKi = 16, aggKd = 8; //Коэффициенты ПИД грубые
double consKp = 4, consKi = 1.5, consKd = 2; //Коэффициенты ПИД точные

PID myPID(&real_I, &angle_I, &ust_I, 0, 0, 0, REVERSE);

void setup(void) {

  Serial.begin(9600);//активируем прием сообщения

  pinMode(detect, INPUT);     //детектор нуля
  pinMode(triac, OUTPUT);      //тиристор

  pinMode(button1, INPUT);    //кнопки
  pinMode(button2, INPUT);

  digitalWrite(button1, HIGH); //включаем подтягивающий резистор
  digitalWrite(button2, HIGH); //включаем подтягивающий резистор

  myPID.SetMode(AUTOMATIC); //Настройка ПИД регулятора
  myPID.SetOutputLimits(1, 625);
  myPID.SetSampleTime(1) ;

  lcd.init();
  lcd.backlight();

  //===========================================================Настройка АЦП

  ADMUX = (0 << REFS1) | (1 << REFS0) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); //

  /*     Биты 7:6 – REFS1:REFS0. Биты выбора опорного напряжения. Если мы будем менять эти биты во время преобразования,
        то изменения вступят в силу только после текущего преобразования. В качестве опорного напряжения может быть выбран AVcc
        (ток источника питания), AREF или внутренний 2.56В источник опорного напряжения.
        Рассмотрим какие же значения можно записать в эти биты:
        REFS1:REFS0
        00    AREF
        01    AVcc, с внешним конденсатором на AREF
        10    Резерв
        11    Внутренний 2.56В  источник, с внешним конденсатором на AREF
        Бит 5 – ADLAR. Определяет как результат преобразования запишется в регистры ADCL и ADCH.
        ADLAR = 0
        Биты 3:0 – MUX3:MUX0 – Биты выбора аналогового канала.
        MUX3:0
        0000      ADC0
        0001      ADC1
        0010      ADC2
        0011      ADC3
        0100      ADC4
        0101      ADC5
        0110      ADC6
        0111      ADC7*/
  //=======================================================================================
  ADCSRA = B11101111; //Включение АЦП
  /*    Бит 7 – ADEN. Разрешение АЦП.
      0 – АЦП выключен
      1 – АЦП включен
      Бит 6 – ADSC. Запуск преобразования (в режиме однократного
      преобразования)
      0 – преобразование завершено
      1 – начать преобразование
      Бит 5 – ADFR. Выбор режима работы АЦП
      0 – режим однократного преобразования
      1 – режим непрерывного преобразования
      Бит 4 – ADIF. Флаг прерывания от АЦП. Бит устанавливается, когда преобразование закончено.
      Бит 3 – ADIE. Разрешение прерывания от АЦП
      0 – прерывание запрещено
      1 – прерывание разрешено
      Прерывание от АЦП генерируется (если разрешено) по завершении преобразования.
      Биты 2:1 – ADPS2:ADPS0. Тактовая частота АЦП
      ADPS2:ADPS0
      000         СК/2
      001         СК/2
      010         СК/4
      011         СК/8
      100         СК/16
      101         СК/32
      110         СК/64
      111         СК/128*/

  //===========================================================
  ACSR = (1 << ACD);
  //------ Timer1 ---------- Таймер задержки времени открытия триака после детектирования нуля (0 триак не откроется)

  TCCR1A = 0x00;  //
  TCCR1B = 0x00;    //
  TCCR1B = (1 << CS12) | (0 << CS11) | (0 << CS10); // Тактирование от CLK.

  // Если нужен предделитель :
  // TCCR1B |= (1<<CS10);           // CLK/1 = 0,0000000625 * 1 = 0,0000000625, 0,01с / 0,0000000625 = 160000 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS11);           // CLK/8 = 0,0000000625 * 8 = 0,0000005, 0,01с / 0,0000005 = 20000 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS10)|(1<<CS11); // CLK/64 = 0,0000000625 * 64 = 0,000004, 0,01с / 0,000004 = 2500 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS12);           // CLK/256  = 0,0000000625 * 256 = 0,000016, 0,01с / 0,000016 = 625 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS10)|(1<<CS12); // CLK/1024
  // Верхняя граница счета. Диапазон от 0 до 255.
  OCR1A = 0;           // Верхняя граница счета. Диапазон от 0 до 65535.

  // Частота прерываний будет = Fclk/(N*(1+OCR1A))
  // где N - коэф. предделителя (1, 8, 64, 256 или 1024)

  TIMSK1 |= (1 << OCIE1A); // Разрешить прерывание по совпадению

  attachInterrupt(1, zero_crosss_int, RISING);//вызов прерывания при детектировании нуля

}
//======================================================== Прерывание по нулю
void zero_crosss_int() {

  PORTD &=  ~(1 << PORTD5);

  TCNT1 = 0;

  zero++;

}

//======================================================== Обработка таймера по совпадению нуля
ISR (TIMER1_COMPA_vect)
{
  PORTD |=  (1 << PORTD5);
}

//======================================================== Обработка прерывания АЦП для расчета среднеквадратичного тока
ISR(ADC_vect)
{
  byte An_pin = ADCL;
  byte An = ADCH;
  Iism = ((An << 8) + An_pin) - 512;
  Iism *= Iism;// возводим значение в квадрат
  Isumm += Iism; // складываем квадраты измерений
  cntr++;
}

void parseParams(String inputString) { //Функция разбора принятой в Serial port строки
  String bb = inputString;
  if (bb.substring(0, 2) == "TW") {
    T1 = bb.substring(bb.indexOf("TW", 2) + 3); //команда
    ust_I = T1.toFloat(); //Выставленный ток с Serial порта
    buffer_ust_I = ust_I;
  }
}

void loop(void) {

  //==================================================================================

  while (Serial.available()) { //Serial port, пока не конец сообщения, читаем данные и формируем строку
    char ch = Serial.read();
    Var += ch;
    if (ch == '\n') {
      parseParams(Var); //вызываем функцию разбора принятой строки
      Var = "";
    }

  }

  if (zero >= 6) {//расчет среднеквадратичного//zero - количество полупериодов дря рассчета среднеквадратичного
    sqrtUsum = Isumm / cntr; //
    real_I = sqrt(sqrtUsum) * 0.029; //
    /*Serial.println(Isumm);
      Serial.println(cntr);
      Serial.println("--------------");*/
    cntr = 0;
    Isumm = 0;// обнуляем суммы напряжений
    zero = 0;
  }

  if (ust_I < 0 || buffer_ust_I < 0) { //Если меньше 0, то 0
    ust_I = 0;
    buffer_ust_I = 0;
  }

  ust_I = (buffer_ust_I > 0) ? buffer_ust_I : 0;

  //=========================================================================Кнопки

  if (digitalRead(button1) == LOW && button1_press == false) { //считаем сколько времени нажата 1 кнопка
    if (millis() - time1_press > 1) {
      button1_press_time++;
      time1_press = millis();
    }
  }

  if (digitalRead(button2) == LOW && button2_press == false) { //считаем сколько времени нажата 2 кнопка
    if (millis() - time2_press > 1) {
      button2_press_time++;
      time2_press = millis();
    }
  }

  if (digitalRead(button1) == HIGH) {//очищаем счетчик если кнопка 1 отпущена
    button1_press = false;
    button1_press_time = 0;
  }

  if (digitalRead(button2) == HIGH) {//очищаем счетчик если кнопка 2 отпущена
    button2_press = false;
    button2_press_time = 0;
  }

  if ( button1_press_time >= 100) { //если более хх мс то кнопка 1 нажата, + установки напряжения
    button1_press_time = 0;
    Serial.println("-----------------");
    buffer_ust_I = buffer_ust_I + 0.01;
  }

  if ( button2_press_time >= 100) { //если более хх мс то кнопка 2 нажата, - установки напряжения
    button2_press_time = 0;
    Serial.println("-----------------");
    buffer_ust_I = buffer_ust_I - 0.01;
  }

  //==============================================ПИД регулятор/грубый регулятор тока
  if ((real_I > (ust_I + 0.01) || real_I < (ust_I - 0.01)) && ust_I != 0)
  {
    myPID.SetTunings(aggKp, aggKi, aggKd);
  }

  //==============================================ПИД регулятор/точный регулятор тока
  if (real_I < (ust_I + 0.01) && real_I > (ust_I - 0.01) && ust_I != 0)
  {
    myPID.SetTunings(consKp, consKi, consKd);
  }

  if (ust_I != 0) myPID.Compute();

  if (angle_I >= 625) {
    OCR1A = 625;
    angle_I = 625;
  }

  if (ust_I == 0) {
    angle_I = 0;
  }

  if (ust_I != 0) {
    OCR1A = int(angle_I);
  }
  else
  {
    OCR1A = 0;
  }

  //=========================Дисплей

  if (millis() - displayMillis >= 300) {
    displayMillis = millis();

    lcd.setCursor(0, 0);
    lcd.print("TOK:");
    lcd.print(real_I);

    lcd.setCursor(0, 1);
    lcd.print("UST:");
    lcd.print(ust_I);
  }

  //=========================Информация в порт раз в секунду

  if (millis() - previousMillis >= 1000) {
    previousMillis = millis();

    Serial.print("realI "); //Измеренный среднеквадратичный ток
    Serial.println(real_I);

    Serial.print("ustI  "); //Установленный выходной ток
    Serial.println(ust_I);

    Serial.println("-----------------");
  }
  //========================================================

}







